/*******************************************************************************
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *******************************************************************************/
package org.ow2.orchestra.console.GWTConsole.client;

import com.google.gwt.user.client.rpc.AsyncCallback;
import java.util.List;
import org.ow2.orchestra.console.GWTConsole.client.Commons.FilterInstanceProcess;
import org.ow2.orchestra.console.GWTConsole.client.ListInstance.MInstance;
import org.ow2.orchestra.console.GWTConsole.client.ListInstance.MInstancePage;
import org.ow2.orchestra.console.GWTConsole.client.ListInstanceActivity.MInstanceActivity;
import org.ow2.orchestra.console.GWTConsole.client.ListProcess.MProcess;
import org.ow2.orchestra.console.GWTConsole.client.ListProcessActivity.MProcessActivity;
import org.ow2.orchestra.console.GWTConsole.client.listJobs.MJob;
import org.ow2.orchestra.console.GWTConsole.client.listMessages.MMessage;
import org.ow2.orchestra.console.GWTConsole.client.userTask.MTask;
import org.ow2.orchestra.console.GWTConsole.client.userTask.MTaskDetails;

/**
 * Interface of the OrchestraManagement service
 *
 * @author Mathieu Crassin
 */
public interface OrchestraManagementServicesAsync {

  /**
   * Get the file named filename on the disk and deploy it to orchestra.
   *
   * @param filename : filename of the file to deploy
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  void deployToOrchestra(final String filename, AsyncCallback<Void> callback);

  /**
   * Get the process list deploy on the orchestra server.
   *
   * @return List<MProcess> : The process list.
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void getProcesses(AsyncCallback<List<MProcess>> callback);

  void getProcess(final String processUUID, AsyncCallback<MProcess> callback);

  /**
   * Undeploy a process from the orchestra server thanks to its UUID.
   *
   * @param processUUID : Information about the process to undeploy.
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  void undeployFromOrchestra(final String processUUID, AsyncCallback<Void> callback);

  /**
   * Get the instances list on the orchestra server
   *
   * @param processId : specify it to get the instances for this specific
   *                  process. When null, all instances will be returned.
   * @return List<MInstance> : The list filled with instances
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void getInstances(String processId, AsyncCallback<List<MInstance>> callback);

  /**
   * Get an instance on the orchestra server
   *
   * @param instanceId The instance UUID as string
   * @param callback   An {@link AsyncCallback} used by the caller to get the retrieved {@link MInstance}
   */
  void getInstance(String instanceId, AsyncCallback<MInstance> callback);

  /**
   * Retrieve a page of existing instances in database.
   *
   * @param pageStart   The start index of the page (depend on sortColumn, isAscending and processId parameters)
   * @param pageLength  How many instances should be retrieved within the page
   * @param sortColumn  The column on which a sort applies. If <code>null</code> there is no sort.
   * @param isAscending Specify if the sort should be ascending or descending. Applies only if sortColumn is not
   *                    <code>null</code>.
   * @param processId   If specified only instances of this process are retrieved. If <code>null</code> instances of any
   *                    processes will be retrieved.
   * @param callback    An {@link AsyncCallback} used by the caller to get the retrieved {@link MInstancePage}
   */
  void getInstances(int pageStart, int pageLength, String sortColumn, boolean isAscending, String processId,
                    FilterInstanceProcess filterState, AsyncCallback<MInstancePage> callback);

  /**
   * Get the activities of a given process
   *
   * @param processId : it's the UUID of the process
   * @return List<MProcessActivity> : The list filled with process' activities
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void getProcessActivities(String processId, AsyncCallback<List<MProcessActivity>> callback);

  /**
   * Get the activities of a given instance
   *
   * @param instanceId : it's the UUID of the instance
   * @return List<MInstanceActivity> : The list filled with instance' activities
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void getInstanceActivities(String instanceId, AsyncCallback<List<MInstanceActivity>> callback);

  void getActivity(String activityId, AsyncCallback<MInstanceActivity> callback);

  /**
   * Suspend a running instance.
   *
   * @param instanceId : the id of the instance to suspend
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void suspendInstance(String instanceId, AsyncCallback<Void> callback);

  /**
   * Resume a suspended instance.
   *
   * @param instanceId : the id of the instance to resume
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void resumeInstance(String instanceId, AsyncCallback<Void> callback);

  /**
   * Exit an instance.
   *
   * @param instanceId : the id of the instance to exit
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void exitInstance(String instanceId, AsyncCallback<Void> callback);

  /**
   * Remove an instance.
   *
   * @param instanceId : the id of the instance to remove
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void removeInstance(String instanceId, AsyncCallback<Void> callback);

  /**
   * Remove a process.
   *
   * @param instanceId : the id of the process to remove
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  void removeProcess(String processId, AsyncCallback<Void> callback);

  /**
   * Retire a process.
   *
   * @param instanceId : the id of the process to retire
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Guillaume Porcher
   */
  void retireProcess(String processId, AsyncCallback<Void> callback);

  /**
   * Activate a process.
   *
   * @param instanceId : the id of the process to activate
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Guillaume Porcher
   */
  void activateProcess(String processId, AsyncCallback<Void> callback);

  void getSerializedDesignerModel(String processId, AsyncCallback<String> callback);

  void getJobsWithException(final AsyncCallback<List<MJob>> async);

  void retryJob(MJob job, final AsyncCallback<Void> async);

  void getPendingMessages(AsyncCallback<List<MMessage>> asyncCallback);

  void removePendingMessage(MMessage object, AsyncCallback<Void> asyncCallback);

  // TASK MANAGEMENT
  void isTaskSupported(AsyncCallback<Boolean> async);

  void getTasks(AsyncCallback<List<MTask>> asyncCallback);

  void startTask(String taskId, AsyncCallback<Void> async);

  void completeTask(String taskId, String data, AsyncCallback<Void> async);

  void getTaskDetails(String taskId, AsyncCallback<MTaskDetails> async);

  void releaseTask(String taskId, AsyncCallback<Void> asyncCallback);

  void resumeTask(String taskId, AsyncCallback<Void> asyncCallback);

  void suspendTask(String taskId, AsyncCallback<Void> asyncCallback);

  void stopTask(String taskId, AsyncCallback<Void> asyncCallback);

  void claimTask(String taskId, AsyncCallback<Void> asyncCallback);

}
