/*******************************************************************************
 * Copyright (C) 2011  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *******************************************************************************/
package org.ow2.orchestra.console.GWTConsole.client.ProcessDetails;

import com.google.gwt.canvas.client.Canvas;
import com.google.gwt.canvas.dom.client.CanvasGradient;
import com.google.gwt.canvas.dom.client.CssColor;
import com.google.gwt.canvas.dom.client.FillStrokeStyle;
import com.google.gwt.user.client.ui.Composite;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeSet;
import org.ow2.orchestra.console.GWTConsole.client.ConsoleEventBus;
import org.ow2.orchestra.console.GWTConsole.client.Event.RefreshDoneEvent;
import org.ow2.orchestra.console.GWTConsole.client.ListInstanceActivity.MInstanceActivity;
import org.ow2.orchestra.console.GWTConsole.client.ListProcessActivity.MProcessActivity;
import org.ow2.orchestra.console.GWTConsole.client.ListProcessActivity.MProcessSubActivity;


/**
 * Widget to draw a graphical representation of a process.
 *
 * @author Guillaume Porcher
 */
public class ProcessCanvasViewer extends Composite {

  private final Canvas canvas = Canvas.createIfSupported();

  private Map<String, MProcessActivity> activities;
  private MProcessActivity process;
  private Map<String, String> states = new HashMap<String, String>();


  public ProcessCanvasViewer() {
    this.initWidget(this.canvas);
  }

  public void setData(final List<MProcessActivity> processActivities) {
    this.process = null;
    this.activities = new HashMap<String, MProcessActivity>();

    for (final MProcessActivity activity : processActivities) {
      if (activity.getProcessActivityType().equals("PROCESS")) {
        this.process = activity;
      }
      this.activities.put(activity.getProcessActivityId(), activity);
    }

    if (this.process != null) {
      this.redrawCanvas();
    }
  }

  public void setInstanceData(final List<MInstanceActivity> instanceActivities) {
    this.states = new HashMap<String, String>();

    for (final MInstanceActivity instanceActivity : instanceActivities) {
      this.states.put(instanceActivity.getInstanceActivityDefId(), instanceActivity.getInstanceActivityState());
    }

    if (this.process != null) {
      this.redrawCanvas();
    }
  }

  private void redrawCanvas() {
    final ActivityCanvas result = this.drawCanvas(this.process, this.activities);
    this.drawFlowLinks(result);
    this.canvas.setCoordinateSpaceHeight(result.getHeight());
    this.canvas.setCoordinateSpaceWidth(result.getWidth());
    this.canvas.getContext2d().restore();
    this.canvas.getContext2d().drawImage(result.getCanvas().getCanvasElement(), 0, 0);
    ConsoleEventBus.getEventBus().fireEvent(new RefreshDoneEvent());
  }

  private FillStrokeStyle getFillColor(final String activityName, final Canvas c) {
    return this.getFillColor(activityName, c, c.getCoordinateSpaceWidth(), c.getCoordinateSpaceHeight());
  }

  private FillStrokeStyle getFillColor(final String activityName, final Canvas c, final int width, final int height) {
    final CanvasGradient g =
            c.getContext2d().createLinearGradient(0, 0, width, height);
    if (this.states.containsKey(activityName)) {
      final String state = this.states.get(activityName);
      if (state.equals("RUNNING")) {
        g.addColorStop(0, "#ffff00");
        g.addColorStop(0.7, "#ffffff");
        return g;
      } else if (state.equals("FINISHED")) {
        g.addColorStop(0, "#00ff00");
        g.addColorStop(0.7, "#ffffff");
        return g;
      } else if (state.equals("EXITED")) {
        g.addColorStop(0, "#ff0000");
        g.addColorStop(0.7, "#ffffff");
        return g;
      }
    }
    g.addColorStop(0, "#cccccc");
    g.addColorStop(0.7, "#ffffff");
    return g;
  }

  private ActivityCanvas drawCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    if (currentActivity.getProcessActivityType().equals("FLOW")) {
      return this.drawFlowCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("SEQUENCE")) {
      return this.drawSequenceCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("RECEIVE")) {
      return this.drawReceiveCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("REPLY")) {
      return this.drawReplyCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("INVOKE")) {
      return this.drawInvokeCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("WAIT")) {
      return this.drawWaitCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("ASSIGN")) {
      return this.drawAssignCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("IF")) {
      return this.drawIfCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("WHILE")) {
      return this.drawWhileCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("PICK")) {
      return this.drawPickCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("SCOPE")
        || currentActivity.getProcessActivityType().equals("PROCESS")) {
      return this.drawScopeCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("COMPENSATE")) {
      return this.drawCompensateCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("THROW")
        || currentActivity.getProcessActivityType().equals("RETHROW")) {
      return this.drawThrowCanvas(currentActivity, activities);
    }
    if (currentActivity.getProcessActivityType().equals("peopleActivity")) {
      return this.drawPeopleActivityCanvas(currentActivity, activities);
    }
    final Canvas canvas = Canvas.createIfSupported();
    int verticalPosition = 30;
    int width = 20 + (int) canvas.getContext2d().measureText(
            currentActivity.getProcessActivityName() + " (" + currentActivity.getProcessActivityType() + ")").getWidth();
    int height = 30;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      verticalPosition += 20;
      width = Math.max(width, 20 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    final List<ActivityCanvas> children = new ArrayList<ActivityCanvas>();
    if (currentActivity.getChildrenActivities() != null) {
      for (final String childName : currentActivity.getChildrenActivities()) {
        final ActivityCanvas childCanvas = this.drawCanvas(activities.get(childName), activities);
        children.add(childCanvas);
        verticalPosition += childCanvas.getHeight() + 10;
        width = Math.max(width, 20 + childCanvas.getWidth());
        height = verticalPosition;
      }
    }
    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));


    verticalPosition = 30;

    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName() + " (" + currentActivity.getProcessActivityType() + ")", 10, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 10, 40);
      verticalPosition += 20;
    }

    for (final ActivityCanvas childCanvas : children) {
      canvas.getContext2d().drawImage(childCanvas.getCanvas().getCanvasElement(), 10, verticalPosition);
      childCanvas.setPosX(10);
      childCanvas.setPosY(verticalPosition);
      verticalPosition += childCanvas.getHeight() + 10;
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, children);
  }

  private ActivityCanvas drawScopeCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int width = 20 + (int) canvas.getContext2d().measureText(
            currentActivity.getProcessActivityName()).getWidth();

    final String childName = currentActivity.getChildrenActivities().get(0);
    final ActivityCanvas childCanvas = this.drawCanvas(activities.get(childName), activities);
    width = Math.max(width, 20 + childCanvas.getWidth());
    int height = childCanvas.getHeight() + 40;
    int verticalPosition = 10;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      verticalPosition += 20;
      width = Math.max(width, 20 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }
    // Sub activities
    int subActivityWidth = 0;
    int subActivityHeight = 0;

    final Map<String, ActivityCanvas> subActivities = new HashMap<String, ActivityCanvas>();
    if (currentActivity.getSubActivities() != null) {
      for (final MProcessSubActivity subActivity : currentActivity.getSubActivities()) {
        final MProcessActivity subActivityProcessActivity;
        if (subActivity.getType().equals("COMPENSATION_HANDLER")
            || subActivity.getType().equals("TERMINATION_HANDLER")
            || subActivity.getType().equals("FAULT_HANDLER")) {
          final MProcessActivity fctActivity = activities.get(subActivity.getChildActivity());
          // show only user defined handlers.
          if (!fctActivity.getProcessActivityName().startsWith("autoGenerated_")) {
            subActivityProcessActivity = activities.get(fctActivity.getChildrenActivities().get(0));
          } else {
            subActivityProcessActivity = null;
          }
        } else {
          subActivityProcessActivity = activities.get(subActivity.getChildActivity());
        }
        if (subActivityProcessActivity != null) {
          final ActivityCanvas subActivityCanvas = this.drawCanvas(subActivityProcessActivity, activities);
          subActivities.put(subActivity.getChildActivity(), subActivityCanvas);
          verticalPosition += subActivityCanvas.getHeight() + 10;
          subActivityWidth = Math.max(subActivityWidth, 30 + subActivityCanvas.getWidth());
          subActivityHeight = verticalPosition;
        }
      }
    }

    // End of sub activities
    final int totalHeight = Math.max(height, subActivityHeight);
    final int totalWidth = Math.max(width, 20 + childCanvas.getWidth() + subActivityWidth);
    canvas.setCoordinateSpaceHeight(totalHeight);
    canvas.setCoordinateSpaceWidth(totalWidth);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas, width, totalHeight));
    canvas.getContext2d().fillRect(0, 0, width, totalHeight);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 10, 20);
    canvas.getContext2d().setFont(font);
    verticalPosition = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      verticalPosition = 50;
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 10, 40);
    }
    canvas.getContext2d().drawImage(childCanvas.getCanvas().getCanvasElement(), 10, verticalPosition);
    childCanvas.setPosX(10);
    childCanvas.setPosY(verticalPosition);

    canvas.getContext2d().setLineWidth(3);
    canvas.getContext2d().strokeRect(0, 0, width, totalHeight);

    verticalPosition = 10;
    if (currentActivity.getSubActivities() != null) {
      for (final MProcessSubActivity subActivity : currentActivity.getSubActivities()) {
        final ActivityCanvas subActivityCanvas = subActivities.get(subActivity.getChildActivity());
        if (subActivityCanvas != null) {
          canvas.getContext2d().setLineWidth(1);
          canvas.getContext2d().setFillStyle("#ffffff");
          canvas.getContext2d().setStrokeStyle("#000000");

          canvas.getContext2d().beginPath();
          canvas.getContext2d().arc(width, verticalPosition + 10, 10, 0, 2 * Math.PI);
          canvas.getContext2d().fill();
          canvas.getContext2d().stroke();
          if (subActivity.getType().equals("COMPENSATION_HANDLER")) {
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(width - 1, verticalPosition + 5);
            canvas.getContext2d().lineTo(width - 1, verticalPosition + 15);
            canvas.getContext2d().lineTo(width - 5, verticalPosition + 10);
            canvas.getContext2d().lineTo(width - 1, verticalPosition + 5);
            canvas.getContext2d().fill();
            canvas.getContext2d().stroke();
            canvas.getContext2d().moveTo(width + 5, verticalPosition + 5);
            canvas.getContext2d().lineTo(width + 5, verticalPosition + 15);
            canvas.getContext2d().lineTo(width + 1, verticalPosition + 10);
            canvas.getContext2d().lineTo(width + 5, verticalPosition + 5);
            canvas.getContext2d().fill();
            canvas.getContext2d().stroke();
          } else if (subActivity.getType().equals("FAULT_HANDLER")) {
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(width - 5, verticalPosition + 15);
            canvas.getContext2d().lineTo(width - 2, verticalPosition + 6);
            canvas.getContext2d().lineTo(width + 2, verticalPosition + 12);
            canvas.getContext2d().lineTo(width + 5, verticalPosition + 5);
            canvas.getContext2d().lineTo(width + 2, verticalPosition + 14);
            canvas.getContext2d().lineTo(width - 2, verticalPosition + 8);
            canvas.getContext2d().lineTo(width - 5, verticalPosition + 15);
            canvas.getContext2d().fill();
            canvas.getContext2d().stroke();
          } else if (subActivity.getType().equals("TERMINATION_HANDLER")) {
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(width - 5, verticalPosition + 15);
            canvas.getContext2d().lineTo(width + 5, verticalPosition + 5);
            canvas.getContext2d().stroke();
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(width - 5, verticalPosition + 5);
            canvas.getContext2d().lineTo(width + 5, verticalPosition + 15);
            canvas.getContext2d().stroke();
          } else if (subActivity.getType().equals("ON_ALARM")) {
            canvas.getContext2d().save();
            // draw icon
            canvas.getContext2d().translate(width, verticalPosition + 10);
            canvas.getContext2d().scale(0.8, 0.8);
            canvas.getContext2d().save();
            canvas.getContext2d().beginPath();
            canvas.getContext2d().arc(0, 0, 10, 0, Math.PI * 2, true);
            canvas.getContext2d().closePath();
            canvas.getContext2d().stroke();
            for (int i = 0; i < 12; i++) {
              canvas.getContext2d().beginPath();
              canvas.getContext2d().moveTo(10, 0);
              canvas.getContext2d().lineTo(8, 0);
              canvas.getContext2d().closePath();
              canvas.getContext2d().stroke();
              canvas.getContext2d().rotate(Math.PI / 6);
            }
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(0, 0);
            canvas.getContext2d().lineTo(4, 0);
            canvas.getContext2d().closePath();
            canvas.getContext2d().stroke();
            canvas.getContext2d().rotate(-Math.PI / 1.8);
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(0, 0);
            canvas.getContext2d().lineTo(8, 0);
            canvas.getContext2d().closePath();
            canvas.getContext2d().stroke();

            canvas.getContext2d().restore();
            canvas.getContext2d().restore();
          } else if (subActivity.getType().equals("ON_EVENT")) {
            canvas.getContext2d().save();
            // draw icon
            canvas.getContext2d().translate(width, verticalPosition + 10);
            canvas.getContext2d().strokeRect(-7, -4, 14, 8);
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(-7, -4);
            canvas.getContext2d().lineTo(0, 0);
            canvas.getContext2d().lineTo(7, -4);
            canvas.getContext2d().stroke();

            canvas.getContext2d().restore();

          }
          canvas.getContext2d().setLineWidth(1);
          canvas.getContext2d().setFillStyle("#000000");
          canvas.getContext2d().setStrokeStyle("#000000");
          this.drawArrow(canvas, width + 10,
                  verticalPosition + 10,
                  30 + width,
                  verticalPosition + 10);
          canvas.getContext2d().drawImage(
                  subActivityCanvas.getCanvas().getCanvasElement(),
                  30 + width,
                  verticalPosition);
          subActivityCanvas.setPosX(30 + width);
          subActivityCanvas.setPosY(verticalPosition);
          verticalPosition += subActivityCanvas.getHeight() + 10;
        }
      }
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, totalHeight, totalWidth, Collections.singleton(childCanvas));
  }

  private ActivityCanvas drawThrowCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 40 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 40 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);

    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(10, 20);
    canvas.getContext2d().lineTo(13, 11);
    canvas.getContext2d().lineTo(17, 17);
    canvas.getContext2d().lineTo(20, 10);
    canvas.getContext2d().lineTo(17, 19);
    canvas.getContext2d().lineTo(13, 13);
    canvas.getContext2d().lineTo(10, 20);
    canvas.getContext2d().fill();
    canvas.getContext2d().stroke();
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 30, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 30, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawAssignCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setLineWidth(1);
    canvas.getContext2d().strokeRect(10, 10, 20, 10);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(14, 12);
    canvas.getContext2d().lineTo(26, 12);
    canvas.getContext2d().moveTo(14, 15);
    canvas.getContext2d().lineTo(26, 15);
    canvas.getContext2d().moveTo(14, 18);
    canvas.getContext2d().lineTo(26, 18);
    canvas.getContext2d().stroke();
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawWaitCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }

    // draw icon
    canvas.getContext2d().translate(20, 15);
    canvas.getContext2d().save();
    canvas.getContext2d().beginPath();
    canvas.getContext2d().arc(0, 0, 10, 0, Math.PI * 2, true);
    canvas.getContext2d().closePath();
    canvas.getContext2d().stroke();
    for (int i = 0; i < 12; i++) {
      canvas.getContext2d().beginPath();
      canvas.getContext2d().moveTo(10, 0);
      canvas.getContext2d().lineTo(8, 0);
      canvas.getContext2d().closePath();
      canvas.getContext2d().stroke();
      canvas.getContext2d().rotate(Math.PI / 6);
    }
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(4, 0);
    canvas.getContext2d().closePath();
    canvas.getContext2d().stroke();
    canvas.getContext2d().rotate(-Math.PI / 1.8);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(8, 0);
    canvas.getContext2d().closePath();
    canvas.getContext2d().stroke();
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawInvokeCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }
    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }

    // draw icon
    canvas.getContext2d().setFillStyle(CssColor.make("#000000"));
    canvas.getContext2d().setStrokeStyle(CssColor.make("#000000"));


    canvas.getContext2d().translate(10, 10);
    canvas.getContext2d().scale(2, 2);
    canvas.getContext2d().setFillStyle(CssColor.make("#000000"));
    canvas.getContext2d().setStrokeStyle(CssColor.make("#000000"));
    canvas.getContext2d().setLineWidth(1);
    canvas.getContext2d().save();
    for (int k = 0; k < 2; k++) {
      for (int j = 0; j < 2; j++) {
        canvas.getContext2d().beginPath();
        canvas.getContext2d().arc(0, 0, 1.5, 0, Math.PI * 2, true);
        canvas.getContext2d().closePath();
        canvas.getContext2d().fill();
        final double lw = canvas.getContext2d().getLineWidth();
        canvas.getContext2d().setLineWidth(1.5 * lw);
        canvas.getContext2d().stroke();
        canvas.getContext2d().setLineWidth(lw);
        for (int i = 0; i < 8; i++) {
          canvas.getContext2d().beginPath();
          canvas.getContext2d().moveTo(3, 0);
          canvas.getContext2d().lineTo(2, 0);
          canvas.getContext2d().closePath();
          canvas.getContext2d().stroke();
          canvas.getContext2d().rotate(Math.PI / 4);
        }
        canvas.getContext2d().translate(4, 4);
      }
      canvas.getContext2d().restore();
      canvas.getContext2d().setFillStyle(CssColor.make("#cccccc"));
      canvas.getContext2d().setStrokeStyle(CssColor.make("#cccccc"));
      canvas.getContext2d().setLineWidth(0.5);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawReplyCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().fillRect(10, 10, 20, 10);
    canvas.getContext2d().setStrokeStyle(CssColor.make(255, 255, 255));
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(10, 10);
    canvas.getContext2d().lineTo(20, 15);
    canvas.getContext2d().lineTo(30, 10);
    canvas.getContext2d().stroke();
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawPeopleActivityCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    // draw circle
    canvas.getContext2d().beginPath();
    canvas.getContext2d().arc(15, 10, 5, 0, Math.PI * 2, true);
    canvas.getContext2d().closePath();
    canvas.getContext2d().stroke();
    canvas.getContext2d().moveTo(7, 25);
    canvas.getContext2d().quadraticCurveTo(15, 4, 23, 25);
    canvas.getContext2d().stroke();
    // write text
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawFlowCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int horizontalPosition = 10;
    int width = 20 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    final Map<String, String> linkSourceActivities = new HashMap<String, String>();
    final Map<String, Integer> activitiesLevel = new HashMap<String, Integer>();
    final Map<Integer, Integer> levelWidth = new HashMap<Integer, Integer>();
    final Map<Integer, Integer> levelHeight = new HashMap<Integer, Integer>();
    for (final MProcessActivity a : activities.values()) {
      for (final String s : a.getSourceLinks()) {
        linkSourceActivities.put(s, a.getProcessActivityId());
      }
    }
    final Map<String, ActivityCanvas> children = new HashMap<String, ActivityCanvas>();
    if (currentActivity.getChildrenActivities() != null) {
      for (final String childName : currentActivity.getChildrenActivities()) {
        final MProcessActivity child = activities.get(childName);
        final ActivityCanvas childCanvas = this.drawCanvas(child, activities);
        children.put(childName, childCanvas);
        // compute nesting level (flow links from flow start)
        final int level = this.getFlowLevel(linkSourceActivities, activitiesLevel, child, activities);
        if (!levelWidth.containsKey(level)) {
          levelWidth.put(level, 10);
          levelHeight.put(level, 0);
        }
        levelWidth.put(level, levelWidth.get(level) + childCanvas.getWidth() + 10);
        levelHeight.put(level, Math.max(levelHeight.get(level), childCanvas.getHeight()));

      }
    }
    final int maxWidth = new TreeSet<Integer>(levelWidth.values()).last();
    width = Math.max(width, maxWidth);
    int maxHeight = 40;
    for (final int h : levelHeight.values()) {
      maxHeight += h + 20;
    }
    height = Math.max(height, maxHeight);

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 200, 0));
    canvas.getContext2d().setLineWidth(3);
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 10, 20);
    canvas.getContext2d().setFont(font);
    int verticalPosition = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, verticalPosition + 10);
      verticalPosition += 20;
    }
    for (final int level : new TreeSet<Integer>(activitiesLevel.values())) {
      horizontalPosition = 10 + (width - levelWidth.get(level)) / 2;
      for (final Entry<String, ActivityCanvas> childCanvas : children.entrySet()) {
        if (level == activitiesLevel.get(childCanvas.getKey())) {
          canvas.getContext2d().drawImage(
                  childCanvas.getValue().getCanvas().getCanvasElement(),
                  horizontalPosition,
                  verticalPosition);
          childCanvas.getValue().setPosX(horizontalPosition);
          childCanvas.getValue().setPosY(verticalPosition);
          horizontalPosition += childCanvas.getValue().getWidth() + 10;
        }
      }
      verticalPosition += levelHeight.get(level) + 20;
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, children.values());
  }

  private void drawFlowLinks(final ActivityCanvas processCanvas) {
    processCanvas.getCanvas().getContext2d().setLineWidth(1);
    final Map<String, String> linkSourceActivities = new HashMap<String, String>();
    for (final MProcessActivity a : this.activities.values()) {
      for (final String s : a.getSourceLinks()) {
        linkSourceActivities.put(s, a.getProcessActivityId());
      }
    }
    for (final MProcessActivity a : this.activities.values()) {
      for (final String l : a.getTargetLinks()) {
        final ActivityCanvas sourceCanvas = this.getChildCanvas(processCanvas, linkSourceActivities.get(l));
        final ActivityCanvas targetCanvas = this.getChildCanvas(processCanvas, a.getProcessActivityId());
        this.drawArrow(processCanvas.getCanvas(),
                sourceCanvas.getPosX() + sourceCanvas.getWidth() / 2,
                sourceCanvas.getPosY() + sourceCanvas.getHeight(),
                targetCanvas.getPosX() + targetCanvas.getWidth() / 2,
                targetCanvas.getPosY());
      }
    }
  }

  private ActivityCanvas getChildCanvas(final ActivityCanvas parent, final String activityId) {
    if (parent.getActivityId().equals(activityId)) {
      return parent;
    }
    for (final ActivityCanvas child : parent.getChildren()) {
      final ActivityCanvas res = this.getChildCanvas(child, activityId);
      if (res != null) {
        return res;
      }
    }
    return null;
  }

  private int getFlowLevel(
          final Map<String, String> linkSourceActivities,
          final Map<String, Integer> activitiesLevel,
          final MProcessActivity child,
          final Map<String, MProcessActivity> activities) {
    if (!activitiesLevel.containsKey(child.getProcessActivityId())) {
      int level = 0;
      for (final String t : child.getTargetLinks()) {
        level = Math.max(level,
                this.getFlowLevel(linkSourceActivities, activitiesLevel, activities.get(linkSourceActivities.get(t)), activities) + 1);
      }
      if (child.getChildrenActivities() != null) {
        for (final String childName : child.getChildrenActivities()) {
          level = Math.max(level,
                  this.getFlowLevel(linkSourceActivities, activitiesLevel, activities.get(childName), activities));
        }
      }
      activitiesLevel.put(child.getProcessActivityId(), level);
    }
    return activitiesLevel.get(child.getProcessActivityId());
  }

  private ActivityCanvas drawSequenceCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int verticalPosition = 30;
    int width = 20 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      verticalPosition += 20;
    }

    final List<ActivityCanvas> children = new ArrayList<ActivityCanvas>();
    if (currentActivity.getChildrenActivities() != null) {
      for (final String childName : currentActivity.getChildrenActivities()) {
        final ActivityCanvas childCanvas = this.drawCanvas(activities.get(childName), activities);
        children.add(childCanvas);
        verticalPosition += childCanvas.getHeight() + 30;
        width = Math.max(width, 20 + childCanvas.getWidth());
        height = verticalPosition - 20;
      }
    }
    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 200));
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 200));
    canvas.getContext2d().setLineWidth(1);
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 10, 20);
    canvas.getContext2d().setFont(font);

    verticalPosition = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 10, 40);
      verticalPosition += 20;
    }

    int index = 0;
    for (final ActivityCanvas childCanvas : children) {
      canvas.getContext2d().drawImage(
              childCanvas.getCanvas().getCanvasElement(),
              (width - childCanvas.getWidth()) / 2,
              verticalPosition);
      childCanvas.setPosX((width - childCanvas.getWidth()) / 2);
      childCanvas.setPosY(verticalPosition);
      index++;
      verticalPosition += childCanvas.getHeight();
      if (index < children.size()) {
        this.drawArrow(canvas, width / 2, verticalPosition, width / 2, verticalPosition + 30);
        verticalPosition += 30;
      }
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, children);
  }

  private ActivityCanvas drawWhileCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int width = 20 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();

    final String childName = currentActivity.getChildrenActivities().get(0);
    final ActivityCanvas childCanvas = this.drawCanvas(activities.get(childName), activities);
    width = Math.max(width, 30 + childCanvas.getWidth());
    int height = 70 + childCanvas.getHeight();

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 20 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 200));
    canvas.getContext2d().setStrokeStyle(CssColor.make(0, 0, 200));
    canvas.getContext2d().setLineWidth(1);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 10, 20);

    canvas.getContext2d().drawImage(childCanvas.getCanvas().getCanvasElement(),
            (width - childCanvas.getWidth()) / 2, 50);
    childCanvas.setPosX((width - childCanvas.getWidth()) / 2);
    childCanvas.setPosY(50);
    this.drawArrow(canvas, width / 2, 30, width / 2, 50);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(width / 2, 50 + childCanvas.getHeight());
    canvas.getContext2d().lineTo(width / 2, height - 10);
    canvas.getContext2d().lineTo(width - 10, height - 10);
    canvas.getContext2d().lineTo(width - 10, 30);
    canvas.getContext2d().lineTo(width / 2, 30);
    canvas.getContext2d().lineTo(width / 2, 40);
    canvas.getContext2d().stroke();
    canvas.getContext2d().closePath();

    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, Collections.singleton(childCanvas));
  }

  private ActivityCanvas drawIfCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int horizontalPosition = 10;
    int width = 40 + 2 * (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 100;
    int verticalPosition = 100;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      verticalPosition += 20;
      width = Math.max(width, 40 + 2 * (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    final Map<String, ActivityCanvas> children = new HashMap<String, ActivityCanvas>();
    if (currentActivity.getChildrenActivities() != null) {
      for (final String childName : currentActivity.getChildrenActivities()) {
        final MProcessActivity child = activities.get(childName);
        final ActivityCanvas childCanvas = this.drawCanvas(child, activities);
        children.put(childName, childCanvas);
        horizontalPosition += childCanvas.getWidth() + 10;
        width = Math.max(width, horizontalPosition);
        height = Math.max(height, verticalPosition + childCanvas.getHeight());
      }
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), width / 2 + 20, 20);
    canvas.getContext2d().setFont(font);

    verticalPosition = 50;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), width / 2 + 20, 40);
      verticalPosition += 20;
    }
    canvas.getContext2d().setStrokeStyle(CssColor.make("#00c800"));
    canvas.getContext2d().setFillStyle(CssColor.make("#00c800"));

    horizontalPosition = 10;
    for (final Entry<String, ActivityCanvas> childCanvas : children.entrySet()) {
      this.drawArrow(canvas, width / 2, verticalPosition - 30, horizontalPosition + childCanvas.getValue().getWidth() / 2, verticalPosition);
      canvas.getContext2d().drawImage(
              childCanvas.getValue().getCanvas().getCanvasElement(),
              horizontalPosition,
              verticalPosition);
      childCanvas.getValue().setPosX(horizontalPosition);
      childCanvas.getValue().setPosY(verticalPosition);
      this.drawArrow(canvas, horizontalPosition + childCanvas.getValue().getWidth() / 2, verticalPosition + childCanvas.getValue().getHeight(), width / 2, height - 20);

      horizontalPosition += childCanvas.getValue().getWidth() + 10;
    }
    final double squareLength = 20 / Math.sqrt(2);
    final CanvasGradient g = canvas.getContext2d().createLinearGradient(0, 0, squareLength, 0);
    g.addColorStop(0, "#00c800");
    g.addColorStop(1, "#ffffff");
    canvas.getContext2d().setFillStyle(g);
    canvas.getContext2d().save();
    canvas.getContext2d().translate(width / 2, 0);
    canvas.getContext2d().rotate(Math.PI / 4);
    canvas.getContext2d().strokeRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().fillRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().restore();
    canvas.getContext2d().save();
    canvas.getContext2d().translate(width / 2, height - 20);
    canvas.getContext2d().rotate(Math.PI / 4);
    canvas.getContext2d().strokeRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().fillRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().restore();
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, children.values());
  }

  private ActivityCanvas drawPickCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {
    final Canvas canvas = Canvas.createIfSupported();
    int horizontalPosition = 10;
    int width = 40 + 2 * (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 100;

    int verticalPosition = 140;

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      verticalPosition += 20;
      width = Math.max(width, 40 + 2 * (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    final Map<String, ActivityCanvas> children = new HashMap<String, ActivityCanvas>();
    if (currentActivity.getSubActivities() != null) {
      for (final MProcessSubActivity subActivity : currentActivity.getSubActivities()) {
        final String childName = subActivity.getChildActivity();
        final MProcessActivity child = activities.get(childName);
        final ActivityCanvas childCanvas = this.drawCanvas(child, activities);
        children.put(childName, childCanvas);
        horizontalPosition += childCanvas.getWidth() + 10;
        width = Math.max(width, horizontalPosition);
        height = Math.max(height, verticalPosition + childCanvas.getHeight());
      }
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));

    verticalPosition = 50;

    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), width / 2 + 20, 20);
    canvas.getContext2d().setFont(font);

    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      verticalPosition += 20;
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), width / 2 + 20, 40);
    }

    canvas.getContext2d().setStrokeStyle(CssColor.make("#00c800"));
    canvas.getContext2d().setFillStyle(CssColor.make("#00c800"));

    horizontalPosition = 10;
    if (currentActivity.getSubActivities() != null) {
      for (final MProcessSubActivity subActivity : currentActivity.getSubActivities()) {
        final ActivityCanvas childCanvas = children.get(subActivity.getChildActivity());
        this.drawArrow(canvas, width / 2, verticalPosition - 30, horizontalPosition + childCanvas.getWidth() / 2, verticalPosition);
        canvas.getContext2d().setStrokeStyle(CssColor.make("#000000"));
        canvas.getContext2d().setFillStyle(this.getFillColor(subActivity.getChildActivity(), canvas, 20, 20));
        if (subActivity.getType().equals("ON_ALARM")) {
          canvas.getContext2d().save();
          // draw icon
          canvas.getContext2d().translate(horizontalPosition + childCanvas.getWidth() / 2, verticalPosition + 10);
          canvas.getContext2d().beginPath();
          canvas.getContext2d().arc(0, 0, 10, 0, Math.PI * 2, true);
          canvas.getContext2d().closePath();
          canvas.getContext2d().fill();
          canvas.getContext2d().stroke();
          canvas.getContext2d().scale(0.8, 0.8);
          canvas.getContext2d().save();
          canvas.getContext2d().beginPath();
          canvas.getContext2d().arc(0, 0, 10, 0, Math.PI * 2, true);
          canvas.getContext2d().closePath();
          canvas.getContext2d().stroke();
          for (int i = 0; i < 12; i++) {
            canvas.getContext2d().beginPath();
            canvas.getContext2d().moveTo(10, 0);
            canvas.getContext2d().lineTo(8, 0);
            canvas.getContext2d().closePath();
            canvas.getContext2d().stroke();
            canvas.getContext2d().rotate(Math.PI / 6);
          }
          canvas.getContext2d().beginPath();
          canvas.getContext2d().moveTo(0, 0);
          canvas.getContext2d().lineTo(4, 0);
          canvas.getContext2d().closePath();
          canvas.getContext2d().stroke();
          canvas.getContext2d().rotate(-Math.PI / 1.8);
          canvas.getContext2d().beginPath();
          canvas.getContext2d().moveTo(0, 0);
          canvas.getContext2d().lineTo(8, 0);
          canvas.getContext2d().closePath();
          canvas.getContext2d().stroke();

          canvas.getContext2d().restore();
          canvas.getContext2d().restore();
        } else if (subActivity.getType().equals("ON_MESSAGE")) {
          canvas.getContext2d().save();
          // draw icon
          canvas.getContext2d().translate(horizontalPosition + childCanvas.getWidth() / 2, verticalPosition + 10);
          canvas.getContext2d().beginPath();
          canvas.getContext2d().arc(0, 0, 10, 0, Math.PI * 2, true);
          canvas.getContext2d().closePath();
          canvas.getContext2d().fill();
          canvas.getContext2d().stroke();
          canvas.getContext2d().strokeRect(-7, -4, 14, 8);
          canvas.getContext2d().beginPath();
          canvas.getContext2d().moveTo(-7, -4);
          canvas.getContext2d().lineTo(0, 0);
          canvas.getContext2d().lineTo(7, -4);
          canvas.getContext2d().stroke();

          canvas.getContext2d().restore();

        }
        canvas.getContext2d().setStrokeStyle(CssColor.make("#000000"));
        canvas.getContext2d().setFillStyle(CssColor.make("#000000"));
        this.drawArrow(canvas, horizontalPosition + childCanvas.getWidth() / 2, verticalPosition + 20, horizontalPosition + childCanvas.getWidth() / 2, verticalPosition + 40);

        canvas.getContext2d().drawImage(
                childCanvas.getCanvas().getCanvasElement(),
                horizontalPosition,
                90);
        childCanvas.setPosX(horizontalPosition);
        childCanvas.setPosY(90);
        canvas.getContext2d().setStrokeStyle(CssColor.make("#00c800"));
        canvas.getContext2d().setFillStyle(CssColor.make("#00c800"));
        this.drawArrow(canvas, horizontalPosition + childCanvas.getWidth() / 2, verticalPosition + 40 + childCanvas.getHeight(), width / 2, height - 20);

        horizontalPosition += childCanvas.getWidth() + 10;
      }
    }
    final double squareLength = 20 / Math.sqrt(2);
    final CanvasGradient g = canvas.getContext2d().createLinearGradient(0, 0, squareLength, 0);
    g.addColorStop(0, "#00c800");
    g.addColorStop(1, "#ffffff");
    canvas.getContext2d().setFillStyle(g);
    // Draw top diamond
    canvas.getContext2d().setStrokeStyle(CssColor.make("#000000"));

    canvas.getContext2d().save();
    canvas.getContext2d().translate(width / 2, 0);
    canvas.getContext2d().rotate(Math.PI / 4);
    canvas.getContext2d().strokeRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().fillRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().restore();
    canvas.getContext2d().save();

    // Draw inline circle
    canvas.getContext2d().translate(width / 2, 10);
    canvas.getContext2d().arc(0, 0, 0.4 * squareLength, 0, 2 * Math.PI);
    canvas.getContext2d().stroke();
    canvas.getContext2d().save();
    // Draw pentagon
    canvas.getContext2d().rotate(0);
    canvas.getContext2d().translate(0, -6);
    canvas.getContext2d().scale(0.6 / Math.sqrt(2), 0.6 / Math.sqrt(2));
    canvas.getContext2d().rotate(54 * Math.PI / 180);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(0, 7.05342302748);
    canvas.getContext2d().stroke();
    canvas.getContext2d().restore();
    canvas.getContext2d().save();
    canvas.getContext2d().rotate(72 * Math.PI / 180);
    canvas.getContext2d().translate(0, -6);
    canvas.getContext2d().rotate(54 * Math.PI / 180);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(0, 7.05342302748);
    canvas.getContext2d().stroke();
    canvas.getContext2d().restore();
    canvas.getContext2d().save();
    canvas.getContext2d().rotate(2 * 72 * Math.PI / 180);
    canvas.getContext2d().translate(0, -6);
    canvas.getContext2d().rotate(54 * Math.PI / 180);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(0, 7.05342302748);
    canvas.getContext2d().stroke();
    canvas.getContext2d().restore();
    canvas.getContext2d().save();
    canvas.getContext2d().rotate(3 * 72 * Math.PI / 180);
    canvas.getContext2d().translate(0, -6);
    canvas.getContext2d().rotate(54 * Math.PI / 180);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(0, 7.05342302748);
    canvas.getContext2d().stroke();
    canvas.getContext2d().restore();
    canvas.getContext2d().save();
    canvas.getContext2d().rotate(4 * 72 * Math.PI / 180);
    canvas.getContext2d().translate(0, -6);
    canvas.getContext2d().rotate(54 * Math.PI / 180);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(0, 0);
    canvas.getContext2d().lineTo(0, 7.05342302748);
    canvas.getContext2d().stroke();
    canvas.getContext2d().restore();

    canvas.getContext2d().restore();
    // Draw bottom diamond
    canvas.getContext2d().setStrokeStyle(CssColor.make("#00c800"));

    canvas.getContext2d().save();
    canvas.getContext2d().translate(width / 2, height - 20);
    canvas.getContext2d().rotate(Math.PI / 4);
    canvas.getContext2d().strokeRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().fillRect(0, 0, squareLength, squareLength);
    canvas.getContext2d().restore();
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width, children.values());
  }

  /**
   * Draws an arrow between P1(x1,x2) and P2(x2, y2)
   *
   * @param canvas
   * @param x1
   * @param y1
   * @param x2
   * @param y2
   */
  private void drawArrow(final Canvas canvas, final int x1, final int y1, final int x2, final int y2) {
    canvas.getContext2d().moveTo(x1, y1);
    canvas.getContext2d().lineTo(x2, y2);
    canvas.getContext2d().stroke();
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(x2, y2);
    final double dx = x2 - x1;
    final double dy = y2 - y1;
    final double norm = Math.sqrt(dx * dx + dy * dy);
    final double vectX = dx / norm;
    final double vectY = dy / norm;
    canvas.getContext2d().lineTo(x2 - 10 * vectX + 5 * vectY, y2 - 10 * vectY - 5 * vectX);
    canvas.getContext2d().lineTo(x2 - 10 * vectX - 5 * vectY, y2 - 10 * vectY + 5 * vectX);
    canvas.getContext2d().lineTo(x2, y2);
    canvas.getContext2d().fill();
    canvas.getContext2d().stroke();
    canvas.getContext2d().closePath();
  }

  private ActivityCanvas drawReceiveCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 50 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 50 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(10, 10, 20, 10);
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(10, 10);
    canvas.getContext2d().lineTo(20, 15);
    canvas.getContext2d().lineTo(30, 10);
    canvas.getContext2d().stroke();
    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 40, 20);
    canvas.getContext2d().setFont(font);
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 40, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }

  private ActivityCanvas drawCompensateCanvas(final MProcessActivity currentActivity, final Map<String, MProcessActivity> activities) {

    final Canvas canvas = Canvas.createIfSupported();
    int width = 40 + (int) canvas.getContext2d().measureText(currentActivity.getProcessActivityName()).getWidth();
    int height = 30;
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      height += 20;
      width = Math.max(width, 40 + (int) canvas.getContext2d().measureText(currentActivity.getDocumentation()).getWidth());
    }

    canvas.setCoordinateSpaceHeight(height);
    canvas.setCoordinateSpaceWidth(width);
    canvas.getContext2d().setFillStyle(this.getFillColor(currentActivity.getProcessActivityId(), canvas));
    canvas.getContext2d().fillRect(0, 0, width, height);
    canvas.getContext2d().strokeRect(0, 0, width, height);

    canvas.getContext2d().setFillStyle(CssColor.make(0, 0, 0));
    canvas.getContext2d().beginPath();
    canvas.getContext2d().moveTo(15, 10);
    canvas.getContext2d().lineTo(15, 20);
    canvas.getContext2d().lineTo(10, 15);
    canvas.getContext2d().lineTo(15, 10);
    canvas.getContext2d().fill();
    canvas.getContext2d().stroke();
    canvas.getContext2d().moveTo(20, 10);
    canvas.getContext2d().lineTo(20, 20);
    canvas.getContext2d().lineTo(15, 15);
    canvas.getContext2d().lineTo(20, 10);
    canvas.getContext2d().fill();
    canvas.getContext2d().stroke();
    String font = canvas.getContext2d().getFont();
    canvas.getContext2d().setFont("italic " + font);
    canvas.getContext2d().fillText(currentActivity.getProcessActivityName(), 30, 20);
    canvas.getContext2d().setFont(font);
    if (currentActivity.getDocumentation() != null && !"".equals(currentActivity.getDocumentation())) {
      canvas.getContext2d().fillText(currentActivity.getDocumentation(), 30, 40);
    }
    return new ActivityCanvas(currentActivity.getProcessActivityId(), canvas, height, width);
  }
}
