/**
 * Copyright (C) 2011  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.console.GWTConsole.client.listJobs;

import com.google.gwt.cell.client.Cell;
import com.google.gwt.cell.client.CompositeCell;
import com.google.gwt.cell.client.HasCell;
import com.google.gwt.core.client.GWT;
import com.google.gwt.user.cellview.client.Column;
import com.google.gwt.user.cellview.client.ColumnSortEvent.ListHandler;
import com.google.gwt.user.cellview.client.IdentityColumn;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.view.client.ListDataProvider;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import org.ow2.orchestra.common.gwt.utils.client.OrchestraGwtException;
import org.ow2.orchestra.common.gwt.utils.client.ui.popup.Popup;
import org.ow2.orchestra.console.GWTConsole.client.Commons.celltable.ConsoleCellTable;
import org.ow2.orchestra.console.GWTConsole.client.Commons.celltable.ImageActionCell;
import org.ow2.orchestra.console.GWTConsole.client.Commons.celltable.LabelCell;
import org.ow2.orchestra.console.GWTConsole.client.Commons.celltable.NullableDateCell;
import org.ow2.orchestra.console.GWTConsole.client.ConsoleEventBus;
import org.ow2.orchestra.console.GWTConsole.client.Event.RefreshDoneEvent;
import org.ow2.orchestra.console.GWTConsole.client.Event.SuspendTimerEvent;
import org.ow2.orchestra.console.GWTConsole.client.Event.ViewRefreshEvent;
import org.ow2.orchestra.console.GWTConsole.client.Event.ViewRefreshHandler;
import org.ow2.orchestra.console.GWTConsole.client.OrchestraManagementServices;
import org.ow2.orchestra.console.GWTConsole.client.OrchestraManagementServicesAsync;
import org.ow2.orchestra.console.GWTConsole.client.Resources.ImageResources;
import org.ow2.orchestra.console.GWTConsole.client.Tab.PTab;
import org.ow2.orchestra.console.GWTConsole.client.Tab.PTab.Type;


/**
 * @author Guillaume Porcher
 *
 */
public class PListJobs {


  protected OrchestraManagementServicesAsync orchestramanagement = GWT.create(OrchestraManagementServices.class);

  // Create a data provider.
  private final ListDataProvider<MJob> dataProvider = new ListDataProvider<MJob>();

  private final String debugPrefix = "listJobs";

  public PListJobs(final VListJobs view) {
    //
    this.addViewRefreshHandlerToBus();

    view.getPager().setDisplay(view.getTable());
    view.getPager().ensureDebugId(debugPrefix + "Pager");

    //
    GWT.log("PListJobs - Constructor : PListJobs correctly instantiated");


    view.getTable().ensureDebugId(debugPrefix);

    // Add a ColumnSortEvent.ListHandler to connect sorting to the
    // java.util.List.
    final ListHandler<MJob> columnSortHandler = new ListHandler<MJob>(this.dataProvider.getList());
    view.getTable().addColumnSortHandler(columnSortHandler);

    this.setColumns(view.getTable(), columnSortHandler);

    // Connect the table to the data provider.
    this.dataProvider.addDataDisplay(view.getTable());

    // Fetch data
    this.refresh();

  }

  private void setColumns(final ConsoleCellTable<MJob> table, final ListHandler<MJob> columnSortHandler) {
    /* ID column */
    final Column<MJob, String> idColumn = new Column<MJob, String>(new LabelCell(this.debugPrefix)) {
      @Override
      public String getValue(final MJob mJob) {
        return Long.toString(mJob.getJobId());
      }
    };
    // Add the columns.
    table.addSortedColumn(columnSortHandler, idColumn, "ID");

    /* Type column */
    final Column<MJob, String> typeColumn = new Column<MJob, String>(new LabelCell(this.debugPrefix)) {
      @Override
      public String getValue(final MJob mJob) {
        if (mJob.getJobType().contains(".")) {
          // Get only simple class name
          return mJob.getJobType().substring(mJob.getJobType().lastIndexOf('.') + 1);
        }
        return mJob.getJobType();
      }
    };
    // Add the columns.
    table.addSortedColumn(columnSortHandler, typeColumn, "Type");

    /* Due date column */
    final Column<MJob, Date> dueDateColumn = new Column<MJob, Date>(new NullableDateCell(this.debugPrefix)) {
      @Override
      public Date getValue(final MJob mJob) {
        return mJob.getDueDate();
      }
    };
    // Add the columns.
    table.addSortedColumn(columnSortHandler, dueDateColumn, "Due Date");

    /* Exception column */
    final Column<MJob, String> exceptionColumn = new Column<MJob, String>(new LabelCell(true, true, this.debugPrefix)) {
      @Override
      public String getValue(final MJob mJob) {
        return mJob.getException();
      }
    };
    // Add the columns.
    table.addSortedColumn(columnSortHandler, exceptionColumn, "Exception");

    /* Actions column */
    final List<HasCell<MJob, ?>> actionList = new ArrayList<HasCell<MJob,?>>();
    actionList.add(new IdentityColumn<MJob>(this.getRetryActionCell()));
    actionList.add(new IdentityColumn<MJob>(this.getExitActionCell()));

    final CompositeCell<MJob> actionCells = new CompositeCell<MJob>(actionList);
    final Column<MJob, MJob> actionsColumn = new IdentityColumn<MJob>(actionCells);
    actionsColumn.setSortable(false);
    // Add the columns.
    table.addColumn(actionsColumn, "Actions");
  }

  private Cell<MJob> getRetryActionCell() {
    final Cell<MJob> retryAction =
      new ImageActionCell<MJob>(ImageResources.INSTANCE.resume(), ImageResources.INSTANCE.resumeDisabled(), "retry", this.debugPrefix + "-retry") {

        @Override
        public void execute(final MJob object) {
          if (this.isActive(object)) {
            orchestramanagement.retryJob(object, new AsyncCallback<Void>() {

              public void onSuccess(final Void arg0) {
                ConsoleEventBus.getEventBus().fireEvent(new ViewRefreshEvent(object.getProcessInstanceUUID(),
                        Type.LIST_JOBS, true));
              }

              public void onFailure(final Throwable arg0) {
                if (arg0 instanceof OrchestraGwtException) {
                  Popup.show((OrchestraGwtException) arg0);
                } else {
                  Popup.showError("Fail while retry job !", arg0.getMessage());
                }
              }
            });
          }
        }

        @Override
        public boolean isActive(final MJob value) {
          return value.getRetries() == 0;
        }
      };
    return retryAction;
  }

  private Cell<MJob> getExitActionCell() {
    final Cell<MJob> exitAction =
            new ImageActionCell<MJob>(
                    ImageResources.INSTANCE.exit(), ImageResources.INSTANCE.exitDisabled(), "exit", this.debugPrefix + "-exit") {

              @Override
              public void execute(final MJob object) {
                if (this.isActive(object)) {
                  orchestramanagement.exitInstance(object.getProcessInstanceUUID(), new AsyncCallback<Void>() {

                    public void onSuccess(final Void arg0) {
                      ConsoleEventBus.getEventBus().fireEvent(new ViewRefreshEvent(object.getProcessInstanceUUID(),
                                Type.LIST_JOBS, true));
                    }

                    public void onFailure(final Throwable arg0) {
                      if (arg0 instanceof OrchestraGwtException) {
                        Popup.show((OrchestraGwtException) arg0);
                      } else {
                        Popup.showError("Fail while exit instance !", arg0.getMessage());
                      }
                    }
                  });
                }
              }

              @Override
              public boolean isActive(final MJob value) {
                return true;
              }
            };
    return exitAction;
  }

  public void refresh() {
    this.orchestramanagement.getJobsWithException(new AsyncCallback<List<MJob>>() {
      public void onSuccess(final List<MJob> jobs) {
        PListJobs.this.dataProvider.getList().clear();
        PListJobs.this.dataProvider.getList().addAll(jobs);
        ConsoleEventBus.getEventBus().fireEvent(new RefreshDoneEvent());
      }

      public void onFailure(final Throwable caught) {
        GWT.log("JobsCellTable - refreshCurrentList: onFailure : " + caught.getMessage());
        ConsoleEventBus.getEventBus().fireEvent(new SuspendTimerEvent(true));
        if (caught instanceof OrchestraGwtException) {
          Popup.show((OrchestraGwtException) caught);
        } else {
          Popup.showError("Unable to access to Orchestra !", caught.getMessage());
        }
      }
    });
  }

  /**
   * Wait refresh event on the eventBus to launch refreshCurrentList
   *
   * @author Jean-Yves Goubet
   */
  private void addViewRefreshHandlerToBus() {
    ConsoleEventBus.getEventBus().addHandler(ViewRefreshEvent.TYPE, new ViewRefreshHandler() {

      public void onRefresh(final ViewRefreshEvent viewRefreshEvent) {
        if (viewRefreshEvent.getElementType() == PTab.Type.LIST_JOBS) {
          PListJobs.this.refresh();
        }
      }
    });
  }
}
