/*******************************************************************************
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 *******************************************************************************/
package org.ow2.orchestra.console.GWTConsole.server;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.granite.util.Base64;
import org.ow2.orchestra.common.gwt.utils.client.OrchestraGwtException;
import org.ow2.orchestra.common.gwt.utils.server.OrchestraGwtServlet;
import org.ow2.orchestra.console.GWTConsole.client.Commons.FilterInstanceProcess;
import org.ow2.orchestra.console.GWTConsole.client.ListInstance.MInstance;
import org.ow2.orchestra.console.GWTConsole.client.ListInstance.MInstancePage;
import org.ow2.orchestra.console.GWTConsole.client.ListInstance.PListInstanceCellTable;
import org.ow2.orchestra.console.GWTConsole.client.ListInstanceActivity.MInstanceActivity;
import org.ow2.orchestra.console.GWTConsole.client.ListProcess.MProcess;
import org.ow2.orchestra.console.GWTConsole.client.ListProcessActivity.MProcessActivity;
import org.ow2.orchestra.console.GWTConsole.client.ListProcessActivity.MProcessSubActivity;
import org.ow2.orchestra.console.GWTConsole.client.OrchestraManagementServices;
import org.ow2.orchestra.console.GWTConsole.client.listJobs.MJob;
import org.ow2.orchestra.console.GWTConsole.client.listMessages.MMessage;
import org.ow2.orchestra.console.GWTConsole.client.userTask.MTask;
import org.ow2.orchestra.console.GWTConsole.client.userTask.MTaskDetails;
import org.ow2.orchestra.facade.InstanceManagementAPI;
import org.ow2.orchestra.facade.ManagementAPI;
import org.ow2.orchestra.facade.QueryDefinitionAPI;
import org.ow2.orchestra.facade.QueryRuntimeAPI;
import org.ow2.orchestra.facade.criteria.Criteria;
import org.ow2.orchestra.facade.criteria.Order;
import org.ow2.orchestra.facade.criteria.PagedResult;
import org.ow2.orchestra.facade.criteria.ProcessInstancesFiltersFields;
import org.ow2.orchestra.facade.criteria.ProcessInstancesSortableFields;
import org.ow2.orchestra.facade.criteria.Restriction;
import org.ow2.orchestra.facade.data.runtime.JobData;
import org.ow2.orchestra.facade.data.runtime.PendingMessageData;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.BaseFCTDefinition;
import org.ow2.orchestra.facade.def.BpelActivityDefinition;
import org.ow2.orchestra.facade.def.BpelActivityWithChildrenDefinition;
import org.ow2.orchestra.facade.def.BpelActivityWithSingleChildDefinition;
import org.ow2.orchestra.facade.def.ElseIfDefinition;
import org.ow2.orchestra.facade.def.ExtensionActivityDefinition;
import org.ow2.orchestra.facade.def.IfActivityDefinition;
import org.ow2.orchestra.facade.def.OnAlarmDefinition;
import org.ow2.orchestra.facade.def.OnEventDefinition;
import org.ow2.orchestra.facade.def.OnMessageDefinition;
import org.ow2.orchestra.facade.def.PickActivityDefinition;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.ScopeActivityDefinition;
import org.ow2.orchestra.facade.def.Source;
import org.ow2.orchestra.facade.def.Target;
import org.ow2.orchestra.facade.exception.ActivityInstanceNotFoundException;
import org.ow2.orchestra.facade.exception.ActivityNotFoundException;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.JobNotFoundException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.Import;
import org.ow2.orchestra.jaxb.bpmn.JaxbUtil;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.ws_ht.TOrganizationalEntity;
import org.ow2.orchestra.ws_ht.api.TTask;
import org.ow2.orchestra.ws_ht.api.wsdl.TaskOperations;

/**
 * Server that manage the relation between the client and the orchestra server.
 *
 * @author Mathieu Crassin
 */
@SuppressWarnings("serial")
public class OrchestraManagementServlet extends OrchestraGwtServlet
    implements OrchestraManagementServices {
  private static final Logger LOG = Logger.getLogger(OrchestraManagementServlet.class.getName());

  /**
   * Get the file named filename on the disk and deploy it to orchestra.
   *
   * @param filename name of the file to deploy
   * @return int return whether or not the function fail
   *         -1 <-> Error
   *         0  <-> Success
   * @throws OrchestraGwtException : to get some orchestra exception and transmit
   *                                      them to the client side
   * @author mathieu
   */
  public final void deployToOrchestra(final String filename) throws OrchestraGwtException {
    Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "deployToOrchestra: RPC call start");

    // we get the data of the file
    final byte[] buffer = this.getContentFile(filename);

    // then we deploy the data
    ProcessDefinition process = null;
    try {
      // deploy these data to orchestra
      process = this.getAPI(ManagementAPI.class).deployBar(buffer);

    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - deployToOchestra :"
          + "Unable to deploy file on orchestra server !\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Unable to deploy file on orchestra server !",
          "OrchestraManagement - deployToOchestra : \n" + Misc.getStackTraceFrom(ex));
    }

    // check the result
    if (process != null) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "deployToOrchestra: File " + filename + " deployed");
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "name: " + process.getName());
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "deployed date: " + process.getDeployedDate());
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - deployToOchestra : Fail to deploy " + filename);
      throw new OrchestraGwtException("Error while deploy to orchestra server !",
          "OrchestraManagement - deployToOrchestra : Fail to deploy " + filename);
    }
  }

  /**
   * Undeploy a process from the orchestra server thanks to its UUID.
   *
   * @param processUUID : Information about the process to undeploy.
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  public final void undeployFromOrchestra(final String processUUID) throws OrchestraGwtException {
    // then modify the input parameter.
    final ProcessDefinitionUUID uuid = new ProcessDefinitionUUID(processUUID);

    // then, call the function to undeploy the process
    boolean res = false;
    try {
      res = this.getAPI(ManagementAPI.class).undeploy(uuid);

    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - undeployFromOchestra :"
          + "Fail to undeploy process with uuid = " + processUUID + "\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Fail to undeploy process with uuid = " + processUUID,
          "OrchestraManagement - undeployFromOrchestra : \n" + Misc.getStackTraceFrom(ex));
    }
    // check return result
    if (res) {
      Misc.fastDynamicLog(
          OrchestraManagementServlet.LOG, Level.INFO, "OrchestraManagement - undeployFromOchestra : File with uuid = "
          + processUUID + " undeployed");
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - undeployFromOchestra :"
          + "Fail to undeploy process with uuid = " + processUUID);
      throw new OrchestraGwtException("Fail to undeploy process with uuid = " + processUUID,
          "OrchestraManagement - undeployFromOrchestra : Fail to undeploy process with uuid = " + processUUID);
    }
  }

  /**
   * Retires a process from the orchestra server thanks to its UUID.
   *
   * @param processUUID : Information about the process to retire.
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Guillaume Porcher
   */
  public final void retireProcess(final String processUUID) throws OrchestraGwtException {
    // then modify the input parameter.
    final ProcessDefinitionUUID uuid = new ProcessDefinitionUUID(processUUID);

    // then, call the function to undeploy the process
    boolean res = false;
    try {
      res = this.getAPI(ManagementAPI.class).retire(uuid);

    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - retireProcess :"
          + "Fail to retire process with uuid = " + processUUID + "\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Fail to retire process with uuid = " + processUUID,
          "OrchestraManagement - retireProcess : \n" + Misc.getStackTraceFrom(ex));
    }
    // check return result
    if (res) {
      Misc.fastDynamicLog(
          OrchestraManagementServlet.LOG, Level.INFO, "OrchestraManagement - retireProcess : Process with uuid = " + processUUID
          + " retired");
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - retireProcess :"
          + "Fail to retire process with uuid = " + processUUID);
      throw new OrchestraGwtException("Fail to retire process with uuid = " + processUUID,
          "OrchestraManagement - retireProcess : Fail to retire process with uuid = " + processUUID);
    }
  }

  /**
   * Activates a process from the orchestra server thanks to its UUID.
   *
   * @param processUUID : Information about the process to activate.
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Guillaume Porcher
   */
  public final void activateProcess(final String processUUID) throws OrchestraGwtException {
    // then modify the input parameter.
    final ProcessDefinitionUUID uuid = new ProcessDefinitionUUID(processUUID);

    // then, call the function to undeploy the process
    boolean res = false;
    try {
      res = this.getAPI(ManagementAPI.class).activate(uuid);

    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - activateProcess :"
          + "Fail to activate process with uuid = " + processUUID + "\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Fail to activate process with uuid = " + processUUID,
          "OrchestraManagement - activateProcess : \n" + Misc.getStackTraceFrom(ex));
    }
    // check return result
    if (res) {
      Misc.fastDynamicLog(
          OrchestraManagementServlet.LOG, Level.INFO, "OrchestraManagement - activateProcess : Process with uuid = "
          + processUUID + " activated");
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - activateProcess :"
          + "Fail to activate process with uuid = " + processUUID);
      throw new OrchestraGwtException("Fail to activate process with uuid = " + processUUID,
          "OrchestraManagement - activateProcess : Fail to activate process with uuid = " + processUUID);
    }
  }

  /**
   * This method gets the processes on orchestra.
   *
   * @return mProcessList : the list of processes
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final List<MProcess> getProcesses() throws OrchestraGwtException {
    final List<MProcess> mProcessList = new ArrayList<MProcess>();
    List<ProcessDefinition> processDefinitionList = null;

    // get the process list
    try {
      processDefinitionList = this.getAPI(QueryDefinitionAPI.class).findProcessDefinitions();
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getProcesses : \n" + Misc.getStackTraceFrom(ex));
    }
    if (processDefinitionList != null) {
      for (final ProcessDefinition p : processDefinitionList) {
        mProcessList.add(new MProcess(p.getUUID().toString(), p.getName(), p.getNameSpace(), p.getVersion(), p.getDeployedDate(),
            p.getUndeployedDate(), p.getState().toString().toLowerCase()));
      }
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getProcesses :"
          + "Unable to get process list on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get process list on orchestra server !",
          "OrchestraManagement - getProcesses : The return result of orchestra is null.");
    }
    return mProcessList;
  }

  /**
   * This method gets the processes on orchestra.
   *
   * @return mProcessList : the list of processes
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final MProcess getProcess(final String processUUID) throws OrchestraGwtException {
    MProcess p = null;
    ProcessDefinition processDef = null;

    // get the process list
    try {
      processDef = this.getAPI(QueryDefinitionAPI.class).getProcessDefinition(new ProcessDefinitionUUID(processUUID));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getProcesses : \n" + Misc.getStackTraceFrom(ex));
    }

    if (processDef != null) {
      p = new MProcess(
          processDef.getUUID().toString(),
          processDef.getName(),
          processDef.getNameSpace(),
          processDef.getVersion(),
          processDef.getDeployedDate(),
          processDef.getUndeployedDate(),
          processDef.getState().toString().toLowerCase());
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getProcess :"
          + "Unable to get process information on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get process information on orchestra server !",
          "OrchestraManagement - getProcess : The return result of orchestra is null.");
    }

    return p;
  }

  /**
   * This method gets the instances of process on orchestra.
   *
   * @param processId : the id of the process when specified
   * @return mInstanceList : the list of instances
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final List<MInstance> getInstances(final String processId) throws OrchestraGwtException {
    final List<MInstance> mInstanceList = new ArrayList<MInstance>();
    Set<ProcessInstance> processInstanceSet = null;

    // we ask for all instances
    if (processId == null) {
      // case when processId not specified
      try {
        processInstanceSet = this.getAPI(QueryRuntimeAPI.class).findProcessInstances();
      } catch (final Exception ex) {
        Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instances list !\n");
        throw new OrchestraGwtException("Error while accessing instances list !",
            "OrchestraManagement - getInstances : \n" + Misc.getStackTraceFrom(ex));
      }
    }
    // we ask for only the instances of the specified process
    else {
      try {
        processInstanceSet = this.getAPI(QueryRuntimeAPI.class).findProcessInstances(new ProcessDefinitionUUID(processId));
      } catch (final ProcessNotFoundException e) {
        Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instances list of process "
            + processId +
            " !\nThe process was not found on orchestra");
        throw new OrchestraGwtException("Error while accessing instances list of process " + processId + " !",
            "OrchestraManagement - getInstances : The process was not found on orchestra\n" + Misc.getStackTraceFrom(e));
      } catch (final Exception e) {
        Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instances list of process "
            + processId + " !\n");
        throw new OrchestraGwtException("Error while accessing instances list of process " + processId + " !",
            "OrchestraManagement - getInstances :\n" + Misc.getStackTraceFrom(e));
      }
    }

    // check the result
    if (processInstanceSet != null) {
      // the correspondence list that permit to avoid asking the same things
      // many times to orchestra
      final Map<ProcessDefinitionUUID, String> processNameMap = new HashMap<ProcessDefinitionUUID, String>();

      // create the mInstanceList object from ProcessInstance set
      for (final ProcessInstance p : processInstanceSet) {

        // for each instance, we get the process name (thanks to the process ID)
        final String processName = this.searchProcessName(processNameMap, p.getProcessDefinitionUUID());

        // and we create the MInstance object corresponding to the
        // ProcessInstance object
        mInstanceList.add(new MInstance(p.getProcessInstanceUUID().toString(),
            p.getProcessDefinitionUUID().toString(),
            processName,
            p.getState().toString().toLowerCase(),
            p.getStartedDate(),
            p.getProcessInstanceStateUpdate().get(p.getProcessInstanceStateUpdate().size() - 1).getDate()));
      }
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getInstances :"
          + "Unable to get instances list on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get instances list on orchestra server !",
          "OrchestraManagement - getInstances : The return result of orchestra is null.");
    }

    return mInstanceList;
  }

  /**
   * Retrieve a page of existing instances on Orchestra server.
   *
   * @param pageStart   The start index of the page (depend on sortColumn, isAscending and processId parameters)
   * @param pageLength  How many instances should be retrieved within the page
   * @param sortColumn  The column on which a sort applies. If <code>null</code> there is no sort.
   * @param isAscending Specify if the sort should be ascending or descending. Applies only if sortColumn is not
   *                    <code>null</code>.
   * @param processId   If specified only instances of this process are retrieved. If <code>null</code> instances of any
   *                    processes will be retrieved.
   * @param filterState If specified only instances filter with this state are retrieved.
   * @return the retrieved {@link MInstancePage}
   * @throws OrchestraGwtException
   */
  public MInstancePage getInstances(final int pageStart, final int pageLength, final String sortColumn,
                                    final boolean isAscending, String processId, FilterInstanceProcess filterState)
      throws OrchestraGwtException {
    final MInstancePage mInstancePage = new MInstancePage();
    PagedResult<ProcessInstance> page = null;

    final Criteria<ProcessInstance> criteria = new Criteria<ProcessInstance>().setFirstResult(pageStart).setMaxResults(pageLength);
    if (sortColumn != null && !"".equals(sortColumn)) {
      final Order<ProcessInstance> order;
      if (sortColumn.equalsIgnoreCase(PListInstanceCellTable.COLUMN_LAST_UPDATE_DATE_TITLE)) {
        order = new Order<ProcessInstance>(ProcessInstancesSortableFields.LAST_UPDATE_DATE, isAscending);
      } else if (sortColumn.equalsIgnoreCase(PListInstanceCellTable.COLUMN_START_DATE_TITLE)) {
        order = new Order<ProcessInstance>(ProcessInstancesSortableFields.START_DATE, isAscending);
      } else if (sortColumn.equalsIgnoreCase(PListInstanceCellTable.COLUMN_STATE_TITLE)) {
        order = new Order<ProcessInstance>(ProcessInstancesSortableFields.STATE, isAscending);
      } else {
        throw new IllegalStateException("Unsupported Process instance sortable field.");
      }
      criteria.setOrder(order);
    }
    if (processId != null) {
      criteria.addRestriction(Restriction.eq(ProcessInstancesFiltersFields.PROCESS_ID, new ProcessDefinitionUUID(processId)));
    }
    if (filterState != null) {
      // Add criteria for State if needed
      if (filterState.getState() != null) {
        criteria.addRestriction(Restriction.eq(ProcessInstancesFiltersFields.STATE, filterState.getState()));
      }
      // Add criteria for Start date if needed
      if (filterState.getStartFrom() != null && filterState.getStartTo() != null && (filterState.getStartFrom().before
          (filterState.getStartTo()))) {
        criteria.addRestriction(Restriction.between(ProcessInstancesFiltersFields.START_DATE, filterState.getStartFrom(),
            filterState.getStartTo()));
      }
      if (filterState.getStartFrom() != null && filterState.getStartTo() == null) {
        criteria.addRestriction(Restriction.ge(ProcessInstancesFiltersFields.START_DATE, filterState.getStartFrom()));
      }
      if (filterState.getStartFrom() == null && filterState.getStartTo() != null) {
        criteria.addRestriction(Restriction.le(ProcessInstancesFiltersFields.START_DATE, filterState.getStartTo()));
      }

      // Add criteria for End date if needed
      if (filterState.getEndFrom() != null && filterState.getEndTo() != null && (filterState.getEndFrom().before
          (filterState.getEndTo()))) {
        criteria.addRestriction(Restriction.between(ProcessInstancesFiltersFields.LAST_UPDATE_DATE, filterState.getEndFrom(),
            filterState.getEndTo()));
      }
      if (filterState.getEndFrom() != null && filterState.getEndTo() == null) {
        criteria.addRestriction(Restriction.ge(ProcessInstancesFiltersFields.LAST_UPDATE_DATE, filterState.getEndFrom()));
      }
      if (filterState.getEndFrom() == null && filterState.getEndTo() != null) {
        criteria.addRestriction(Restriction.le(ProcessInstancesFiltersFields.LAST_UPDATE_DATE, filterState.getEndTo()));
      }
    }
    try {
      page = this.getAPI(QueryRuntimeAPI.class).findProcessInstancesPage(criteria);
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instances list !\n");
      throw new OrchestraGwtException("Error while accessing instances list !",
          "OrchestraManagement - getInstances : \n" + Misc.getStackTraceFrom(ex));
    }

    // check the result
    if (page != null) {
      // the correspondence list that permit to avoid asking the same things
      // many times to orchestra
      final Map<ProcessDefinitionUUID, String> processNameMap = new HashMap<ProcessDefinitionUUID, String>();

      // create the mInstanceList object from ProcessInstance set
      for (final ProcessInstance p : page.getElements()) {

        // for each instance, we get the process name (thanks to the process ID)
        final String processName = this.searchProcessName(processNameMap, p.getProcessDefinitionUUID());

        // and we create the MInstance object corresponding to the
        // ProcessInstance object
        mInstancePage.getInstances().add(new MInstance(p.getProcessInstanceUUID().toString(),
            p.getProcessDefinitionUUID().toString(),
            processName,
            p.getState().toString().toLowerCase(),
            p.getStartedDate(),
            p.getProcessInstanceStateUpdate().get(p.getProcessInstanceStateUpdate().size() - 1).getDate()));
      }
      mInstancePage.setTotalInstancesCount(page.getTotalElementsCount());
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getInstances :"
          + "Unable to get instances list on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get instances list on orchestra server !",
          "OrchestraManagement - getInstances : The return result of orchestra is null.");
    }

    return mInstancePage;
  }

  /**
   * This method gets the info on a unique instance
   *
   * @param instanceId : the id of the instance
   * @author Jean-Yves Goubet
   */
  public final MInstance getInstance(final String instanceId) throws OrchestraGwtException {
    MInstance i = null;
    ProcessInstance processInstance = null;

    if (instanceId == null || instanceId.equals("")) {
      return null;
    }

    //
    try {
      processInstance = this.getAPI(QueryRuntimeAPI.class).getProcessInstance(new ProcessInstanceUUID(instanceId));
    } catch (final InstanceNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instance information ( id = "
          + instanceId
          + " ) !\nThe instance was not found on orchestra");
      throw new OrchestraGwtException("Error while accessing instance information ( id = " + instanceId + " ) !",
          "OrchestraManagement - getInstance : The instance was not found on orchestra\n" + Misc.getStackTraceFrom(e));
    } catch (final Exception e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing instance information ( id = "
          + instanceId + " ) !");
      throw new OrchestraGwtException("Error while accessing instance information ( id = " + instanceId + " ) !",
          "OrchestraManagement - getInstance :\n" + Misc.getStackTraceFrom(e));
    }

    //
    if (processInstance != null) {
      final String processName = this.searchProcessName(null, processInstance.getProcessDefinitionUUID());
      i = new MInstance(processInstance.getUUID().toString(),
          processInstance.getProcessDefinitionUUID().toString(), processName,
          processInstance.getState().toString().toLowerCase(), processInstance.getStartedDate(),
          processInstance.getProcessInstanceStateUpdate().get(processInstance.
              getProcessInstanceStateUpdate().size() - 1).getDate());
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getInstance :"
          + "Unable to get instance information on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get instance information on orchestra server !",
          "OrchestraManagement - getInstance : The return result of orchestra is null.");
    }

    return i;
  }

  /**
   * This method gets the info of an activity
   *
   * @param activityId : the id of the activity
   * @return MInstanceActivity : the instance activity info
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final MInstanceActivity getActivity(final String activityId) throws OrchestraGwtException {
    MInstanceActivity instanceActivity = null;
    ActivityInstance activityInstance = null;
    String activityName = null;

    //
    try {
      activityInstance = this.getAPI(QueryRuntimeAPI.class).getActivityInstance(new ActivityInstanceUUID(activityId));
      activityName = this.getAPI(QueryDefinitionAPI.class).getActivityDefinition(activityInstance.getDefinitionUUID()).getName();
    } catch (final ActivityNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getActivity :"
          + "Unable to get activity information on orchestra server !\n");
      throw new OrchestraGwtException("Unable to get activity information on orchestra server !",
          "OrchestraManagement - getActivity : Activity not found !");
    } catch (final ActivityInstanceNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getActivity :"
          + "Unable to get activity information on orchestra server !\nInstance activity not found !");
      throw new OrchestraGwtException("Unable to get activity information on orchestra server !",
          "OrchestraManagement - getActivity : Instance activity not found !");
    } catch (final Exception e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getActivity :"
          + "Unable to get activity information on orchestra server !\nUnknown error !");
      throw new OrchestraGwtException("Unable to get activity information on orchestra server !",
          "OrchestraManagement - getActivity : Unknown error !");
    }

    //
    if (activityName != null) {
      instanceActivity = new MInstanceActivity(
          activityInstance.getUUID().toString(),
          activityInstance.getDefinitionUUID().toString(),
          activityName,
          activityInstance.getType().toString(),
          activityInstance.getState().toString(),
          activityInstance.getStartedDate(),
          activityInstance.getEndedDate());

    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getActivity :"
          + "Unable to get activity information on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get activity information on orchestra server !",
          "OrchestraManagement - getActivity : The return result of orchestra is null.");
    }
    return instanceActivity;
  }

  /**
   * Suspend the instance that have the specified Id on the orchestra server.
   *
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  public final void suspendInstance(final String instanceId) throws OrchestraGwtException {
    try {
      final ProcessInstanceUUID inst = new ProcessInstanceUUID(instanceId);
      this.getAPI(InstanceManagementAPI.class).suspend(inst);
    } catch (final InstanceNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - suspendInstance :"
          + "Error, instance " + instanceId + " not found !\n" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Error while suspending the instance with uuid " + instanceId + " !",
          "OrchestraManagement - suspendInstance : Error, instance not found !\n" + Misc.getStackTraceFrom(e));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - suspendInstance :"
          + "Error while suspending instance : " + instanceId + " !\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Error while suspending the instance with uuid " + instanceId + " !",
          "OrchestraManagement - suspendInstance : \n" + Misc.getStackTraceFrom(ex));
    }
  }

  /**
   * Resume the instance that have the specified Id on the orchestra server.
   *
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  public final void resumeInstance(final String instanceId) throws OrchestraGwtException {
    try {
      final ProcessInstanceUUID inst = new ProcessInstanceUUID(instanceId);
      this.getAPI(InstanceManagementAPI.class).resume(inst);
    } catch (final InstanceNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - resumeInstance :"
          + "Error, instance " + instanceId + " not found !\n" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Error while resuming the instance with uuid " + instanceId + " !",
          "OrchestraManagement - resumeInstance : Error, instance not found !\n" + Misc.getStackTraceFrom(e));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - resumeInstance :"
          + "Error while resuming instance : " + instanceId + " !\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Error while resuming the instance with uuid " + instanceId + " !",
          "OrchestraManagement - resumeInstance : \n" + Misc.getStackTraceFrom(ex));
    }
  }

  /**
   * Exit the instance that have the specified Id on the orchestra server.
   *
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  public final void exitInstance(final String instanceId) throws OrchestraGwtException {
    try {
      this.getAPI(InstanceManagementAPI.class).exit(new ProcessInstanceUUID(instanceId));
    } catch (final InstanceNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - exitInstance :"
          + "Error, instance " + instanceId + " not found !\n" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Error while exiting the instance with uuid " + instanceId + " !",
          "OrchestraManagement - exitInstance : Error, instance not found !\n" + Misc.getStackTraceFrom(e));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - exitInstance :"
          + "Error while exiting instance : " + instanceId + " !\n" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Error while exiting the instance with uuid " + instanceId + " !",
          "OrchestraManagement - exitInstance : \n" + Misc.getStackTraceFrom(ex));
    }
  }

  /**
   * Delete the instance that have the specified Id on the orchestra server.
   *
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Mathieu Crassin
   */
  public void removeInstance(final String instanceId) throws OrchestraGwtException {
    // call the function
    boolean res = false;
    try {
      res = this.getAPI(InstanceManagementAPI.class).remove(new ProcessInstanceUUID(instanceId));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - removeInstance : "
          + "Error while removing the instance with uuid " + instanceId + " !" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Error while removing the instance with id " + instanceId + " !",
          "OrchestraManagement - removeInstance : \n" + Misc.getStackTraceFrom(ex));
    }

    // check the result
    if (!res) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - removeInstance :"
          + "Fail to remove instance with id = " + instanceId + ", return result was wrong !");
      throw new OrchestraGwtException("Fail to remove instance with id = " + instanceId,
          "OrchestraManagement - removeInstance : Fail to remove instance with id = "
              + instanceId + ", return result was wrong !");
    }
  }

  /**
   * Remove definitively a process from the orchestra server.
   *
   * @param processId : Id of the process to remove.
   * @author Mathieu Crassin
   */
  public void removeProcess(final String processId) throws OrchestraGwtException {
    // then call the remove function
    boolean res = false;
    try {
      res = this.getAPI(ManagementAPI.class).remove(new ProcessDefinitionUUID(processId));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - removeProcess : "
          + "Error while removing the process with uuid " + processId + " !" + Misc.getStackTraceFrom(ex));
      throw new OrchestraGwtException("Error while removing the process with uuid " + processId + " !",
          "OrchestraManagement - removeProcess : \n" + Misc.getStackTraceFrom(ex));
    }

    // check the result
    if (!res) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - removeProcess :"
          + "Fail to remove process with id = " + processId + ", return result was wrong !");
      throw new OrchestraGwtException("Fail to remove process with id = " + processId,
          "OrchestraManagement - removeProcess : Fail to remove process with id = " + processId + ", return result was wrong !");
    }
  }

  /**
   * Get the activities of a given process
   *
   * @param processId : it's the UUID of the process
   * @return List<MProcessActivity> : The list filled with process' activities
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final List<MProcessActivity> getProcessActivities(final String processId) throws OrchestraGwtException {
    final List<MProcessActivity> mProcessActivityList = new ArrayList<MProcessActivity>();
    Set<ActivityDefinition> processActivitySet = null;

    // get the instance activities set from orchestra
    try {
      processActivitySet = this.getAPI(QueryDefinitionAPI.class).findActivityDefinitions(new ProcessDefinitionUUID(processId));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getInstanceActivities : \n" + Misc.getStackTraceFrom(ex));
    }

    // fill in the list with appropriate values
    if (processActivitySet != null) {
      for (final ActivityDefinition a : processActivitySet) {
        List<String> children = null;
        List<MProcessSubActivity> subActivities = null;
        if (a instanceof IfActivityDefinition) {
          children = new ArrayList<String>();
          final IfActivityDefinition ifAct = (IfActivityDefinition) a;
          children.add(ifAct.getEnclosedActivityUUID().toString());
          for (final ElseIfDefinition elseIfAct : ifAct.getElseIfDefinitionList()) {
            children.add(elseIfAct.getActivityDefinitionUUID().toString());
          }
          if (ifAct.getElseDefinition() != null) {
            children.add(ifAct.getElseDefinition().getActivityDefinitionUUID().toString());
          }
        } else if (a instanceof PickActivityDefinition) {
          subActivities = new ArrayList<MProcessSubActivity>();
          final PickActivityDefinition pickAct = (PickActivityDefinition) a;
          for (final OnAlarmDefinition onAlarm : pickAct.getOnAlarmDefinitionList()) {
            final MProcessSubActivity child = new MProcessSubActivity("ON_ALARM", onAlarm.getActivityDefinitionUUID().toString());
            subActivities.add(child);
          }
          for (final OnMessageDefinition onMessage : pickAct.getOnMessageDefinitionList()) {
            final MProcessSubActivity child = new MProcessSubActivity("ON_MESSAGE", onMessage.getActivityDefinitionUUID().toString());
            subActivities.add(child);
          }
        } else if (a instanceof ScopeActivityDefinition) {
          subActivities = new ArrayList<MProcessSubActivity>();
          final ScopeActivityDefinition scopeAct = (ScopeActivityDefinition) a;
          for (final OnAlarmDefinition onAlarm : scopeAct.getOnAlarmEventHandlers()) {
            final MProcessSubActivity child = new MProcessSubActivity("ON_ALARM", onAlarm.getActivityDefinitionUUID().toString());
            subActivities.add(child);
          }
          for (final OnEventDefinition onEvent : scopeAct.getOnEventEventHandlers()) {
            final MProcessSubActivity child = new MProcessSubActivity("ON_EVENT", onEvent.getActivityDefinitionUUID().toString());
            subActivities.add(child);
          }
          for (final ActivityDefinitionUUID faultHandlerUUID : scopeAct.getFaultHandlerUUIDs()) {
            final MProcessSubActivity child = new MProcessSubActivity("FAULT_HANDLER", faultHandlerUUID.toString());
            subActivities.add(child);
          }
          final MProcessSubActivity compensationChild = new MProcessSubActivity("COMPENSATION_HANDLER", scopeAct.getCompensationHandlerDefinitionUUID().toString());
          subActivities.add(compensationChild);
          final MProcessSubActivity thChild = new MProcessSubActivity("TERMINATION_HANDLER", scopeAct.getTerminationHandlerDefinitionUUID().toString());
          subActivities.add(thChild);
          children = new ArrayList<String>();
          children.add(((BpelActivityWithSingleChildDefinition) a).getEnclosedActivityUUID().toString());
        } else if (a instanceof BpelActivityWithSingleChildDefinition) {
          children = new ArrayList<String>();
          children.add(((BpelActivityWithSingleChildDefinition) a).getEnclosedActivityUUID().toString());
        } else if (a instanceof BpelActivityWithChildrenDefinition) {
          children = new ArrayList<String>();
          for (final ActivityDefinitionUUID childUUID : ((BpelActivityWithChildrenDefinition) a).getEnclosedActivitiesUUID()) {
            children.add(childUUID.toString());
          }
        } else if (a instanceof BaseFCTDefinition) {
          children = new ArrayList<String>();
          children.add(((BaseFCTDefinition) a).getEnclosedActivityUUID().toString());
        }
        final List<String> sources = new ArrayList<String>();
        final List<String> targets = new ArrayList<String>();
        if (a instanceof BpelActivityDefinition) {
          final BpelActivityDefinition bpelActivityDefinition = (BpelActivityDefinition) a;
          if (bpelActivityDefinition.getSourceList() != null) {
            for (final Source s : bpelActivityDefinition.getSourceList()) {
              sources.add(s.getLinkName());
            }
          }
          if (bpelActivityDefinition.getTargetList() != null) {
            for (final Target t : bpelActivityDefinition.getTargetList()) {
              targets.add(t.getLinkName());
            }
          }
        }
        mProcessActivityList.add(
            new MProcessActivity(
                a.getUUID().toString(),
                a.getName(),
                (a instanceof ExtensionActivityDefinition) ? ((ExtensionActivityDefinition) a).getExtensionActivityType() : a.getType().toString(),
                a.getDocumentation(),
                children,
                sources,
                targets,
                subActivities));
      }
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getProcessActivities :"
          + "Unable to get process activities on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get process activities list on orchestra server !",
          "OrchestraManagement - getProcessActivities : The return result of orchestra is null.");
    }

    return mProcessActivityList;
  }

  /**
   * Get the activities of a given instance
   *
   * @param instanceId : it's the UUID of the instance
   * @return List<MInstanceActivity> : The list filled with instance' activities
   * @throws OrchestraGwtException : to get some orchestra exception and
   *                                      transmit them to the client side
   * @author Jean-Yves Goubet
   */
  public final List<MInstanceActivity> getInstanceActivities(final String instanceId) throws OrchestraGwtException {
    final List<MInstanceActivity> mInstanceActivityList = new ArrayList<MInstanceActivity>();
    Set<ActivityInstance> instanceActivitySet = null;

    if (instanceId == null || instanceId.equals("")) {
      return mInstanceActivityList;
    }

    // get the activity instance set from orchestra
    try {
      instanceActivitySet = this.getAPI(QueryRuntimeAPI.class).findActivityInstances(new ProcessInstanceUUID(instanceId));
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getInstanceActivities : \n" + Misc.getStackTraceFrom(ex));
    }

    // fill in the list with appropriate values
    if (instanceActivitySet != null) {
      String activityName = null;
      for (final ActivityInstance a : instanceActivitySet) {
        try {
          activityName = this.getAPI(QueryDefinitionAPI.class).getActivityDefinition(a.getDefinitionUUID()).getName();
        } catch (final Exception ex) {
          Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
          throw new OrchestraGwtException("Error while sending request to orchestra server !",
              "OrchestraManagement - getInstanceActivities (getActivityName()): \n" + Misc.getStackTraceFrom(ex));
        }
        mInstanceActivityList.add(new MInstanceActivity(
            a.getUUID().toString(),
            a.getDefinitionUUID().toString(),
            activityName,
            a.getType().toString(),
            a.getState().toString(),
            a.getStartedDate(),
            a.getEndedDate()));
      }
    } else {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getInstanceActivities :"
          + "Unable to get instance activities on orchestra server !\nThe return result of orchestra is null.");
      throw new OrchestraGwtException("Unable to get instance activities list on orchestra server !",
          "OrchestraManagement - getInstanceActivities : The return result of orchestra is null.");
    }

    return mInstanceActivityList;
  }

  /**
   * Get the entire content of a file.
   *
   * @param filename
   * @return
   * @throws OrchestraGwtException
   * @author Mathieu Crassin
   */
  protected byte[] getContentFile(final String filename) throws OrchestraGwtException {
    try {
      // create file object
      final File file = new File(filename);
      final byte[] buffer = new byte[(int) file.length()];

      // create FileInputStream object
      final FileInputStream fis = new FileInputStream(file);
      try {
        final int size = fis.read(buffer);
        if (size != file.length()) {
          Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getContentFile :"
              + "The file content does not match the file length");
          throw new OrchestraGwtException("Internal server error while accessing file !",
              "OrchestraManagement - getContentFile : The file content does not match the file length");
        }
      } finally {
        fis.close();
      }
      // display
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "OrchestraManagement - getContentFile : File " + filename
          + " found !");
      // and try to delete it
      if (file.delete()) {
        Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.INFO, "OrchestraManagement - getContentFile : Temporary file "
            + filename + " deleted");
      } else {
        Misc.fastDynamicLog(
            OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getContentFile : Error while deleting file "
            + filename);
        throw new OrchestraGwtException("Internal error during connexion with Orchestra  !",
            "OrchestraManagement - getContentFile : Error while deleting file " + filename);
      }

      return buffer;

    } catch (final NullPointerException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getContentFile :"
          + "The file name is a null pointer !\n" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Internal server error while accessing file !",
          "OrchestraManagement - getContentFile : The file name is a null pointer !\n" + Misc.getStackTraceFrom(e));
    } catch (final FileNotFoundException e) {
      Misc.fastDynamicLog(
          OrchestraManagementServlet.LOG, Level.SEVERE,
          "OrchestraManagement - getContentFile : File not found !\n" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Internal server error while accessing file !",
          "OrchestraManagement - getContentFile : File not found !\n" + Misc.getStackTraceFrom(e));
    } catch (final IOException ioe) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getContentFile :"
          + "Exception while reading the file !\n" + Misc.getStackTraceFrom(ioe));
      throw new OrchestraGwtException("Internal server error while accessing file !",
          "OrchestraManagement - getContentFile : Exception while reading the file !\n" + Misc.getStackTraceFrom(ioe));
    } catch (final SecurityException se) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - getContentFile :"
          + "Exception while deleting file !\n" + Misc.getStackTraceFrom(se));
      throw new OrchestraGwtException("Internal server error while accessing file !",
          "OrchestraManagement - getContentFile : Exception while deleting file !\n" + Misc.getStackTraceFrom(se));
    }
  }


  /**
   * This function return the process name corresponding to an uuid. First the
   * function check if a correspondence is registered in the list, in this case,
   * the response can be directly return in the other case, the response is ask
   * to orchestra server.
   *
   * @param processMap
   * @param uuid
   * @return
   * @throws OrchestraGwtException
   */
  protected String searchProcessName(final Map<ProcessDefinitionUUID, String> processMap,
                                     final ProcessDefinitionUUID uuid) throws OrchestraGwtException {

    if (uuid == null) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing process informations");
      throw new OrchestraGwtException("Error while accessing process informations",
          "OrchestraManagement - searchProcessName : Parameter uuid is null");
    }

    String processName = null;
    if (processMap != null) {
      processName = processMap.get(uuid);
    }

    // if we didn't find a correspondence in the list
    if (processName == null) {

      ProcessDefinition processDef;
      try {
        processDef = this.getAPI(QueryDefinitionAPI.class).getProcessDefinition(uuid);
      } catch (final ProcessNotFoundException e) {
        Misc.fastDynamicLog(
            OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing process (id = " + uuid.toString() + " ) "
            + "The process was not found on orchestra !");
        throw new OrchestraGwtException("Error while accessing process (id = " + uuid.toString() + " )"
            + "The process was not found on orchestra !",
            "OrchestraManagement - searchProcessName : " + Misc.getStackTraceFrom(e));
      } catch (final Exception e) {
        Misc.fastDynamicLog(
            OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing process (id = " + uuid.toString() + " )");
        throw new OrchestraGwtException("Error while accessing process (id = " + uuid.toString() + " )",
            "OrchestraManagement - searchProcessName : " + Misc.getStackTraceFrom(e));
      }

      if (processDef != null) {
        processName = processDef.getName();
      } else {
        Misc.fastDynamicLog(
            OrchestraManagementServlet.LOG, Level.SEVERE, "Error while accessing process (id = " + uuid.toString() + " )");
        throw new OrchestraGwtException("Error while accessing process (id = " + uuid.toString() + " )",
            "OrchestraManagement - searchProcessName : The orchestra returned object is null");
      }

      // add the correspondence in the grid
      if (processMap != null) {
        processMap.put(uuid, processName);
      }
    }
    // if we found a correspondence in the list
    return processName;
  }

  public String getSerializedDesignerModel(final String processId) throws OrchestraGwtException {
    try {
      final byte[] bpmnResource =
          this.getAPI(QueryDefinitionAPI.class).getProcessResource(new ProcessDefinitionUUID(processId), "process.bpmn");
      if (bpmnResource != null) {
        //Create tmp directory
        File processDir = File.createTempFile("BpmnProcess_", ".d");
        processDir.delete();
        processDir.mkdirs();
        try {
          File processModel = new File(processDir, "process.bpmn");
          Misc.write(bpmnResource, processModel);
          final Definitions definitions = JaxbUtil.unmarshalBpmn(bpmnResource);
          for (Import anImport : definitions.getImports()) {
            final byte[] importResource =
                this.getAPI(QueryDefinitionAPI.class).getProcessResource(
                    new ProcessDefinitionUUID(processId), anImport.getLocation());
            File fileImport = new File(processDir, anImport.getLocation());
            Misc.write(importResource, fileImport);
          }

          return Base64.encodeToString(
              DesignerModelUtils.getSerializedDesignerModel(definitions, processModel.getAbsolutePath()), false);
        } finally {
          Misc.deleteDir(processDir);
        }
      }
    } catch (final Exception ex) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "Error while sending request to orchestra server !");
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getProcesses : \n" + Misc.getStackTraceFrom(ex));
    }
    return null;
  }

  public List<MJob> getJobsWithException() throws OrchestraGwtException {
    List<MJob> result = new ArrayList<MJob>();
    List<JobData> jobs = this.getAPI(InstanceManagementAPI.class).findJobsWithException();
    for (JobData job : jobs) {
      MJob mJob = new MJob(
          job.getJobType(),
          job.getJobId(),
          job.isSuspended(),
          job.getDueDate(),
          job.getException(),
          job.getRetries(),
          job.getActivityInstanceUUID().toString(),
          job.getProcessInstanceUUID().toString()
      );
      result.add(mJob);
    }
    return result;
  }

  public void retryJob(MJob job) throws OrchestraGwtException {
    JobData jobData = new JobData(
        job.getJobType(),
        job.getJobId(),
        job.isSuspended(),
        job.getDueDate(),
        job.getException(),
        job.getRetries(),
        new ActivityInstanceUUID(job.getActivityInstanceUUID()),
        new ProcessInstanceUUID(job.getProcessInstanceUUID())
    );
    try {
      this.getAPI(InstanceManagementAPI.class).retryJob(jobData, 1);
    } catch (JobNotFoundException e) {
      Misc.fastDynamicLog(OrchestraManagementServlet.LOG, Level.SEVERE, "OrchestraManagement - retryJob : "
          + "Error while retrying the job with id " + job.getJobId() + " !" + Misc.getStackTraceFrom(e));
      throw new OrchestraGwtException("Error while retrying the job with id " + job.getJobId() + " !",
          "OrchestraManagement - retryJob : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public List<MMessage> getPendingMessages() throws OrchestraGwtException {
    List<MMessage> result = new ArrayList<MMessage>();
    List<PendingMessageData> pendingMessageDataList = this.getAPI(InstanceManagementAPI.class).getPendingMessages();
    for (PendingMessageData pendingMessageData : pendingMessageDataList) {
      MMessage message = new MMessage(
          pendingMessageData.getMessageId(),
          pendingMessageData.getMessage().toString(),
          pendingMessageData.getProcessQName().toString(),
          pendingMessageData.getPortTypeQName().toString(),
          pendingMessageData.getOperationName(),
          pendingMessageData.isOneWay(),
          pendingMessageData.getAddedDate()
      );
      result.add(message);
    }
    return result;
  }

  public void removePendingMessage(final MMessage message) throws OrchestraGwtException {
    PendingMessageData pendingMessageData = new PendingMessageData(
        message.getMessageId(),
        null,
        QName.valueOf(message.getProcessQName()),
        QName.valueOf(message.getPortTypeQName()),
        message.getOperationName(),
        message.isOneWay(),
        message.getAddedDate()
    );
    this.getAPI(InstanceManagementAPI.class).removePendingMessage(pendingMessageData);
  }


  // TASK OPERATIONS
  public boolean isTaskSupported() {
    try {
      return this.getAPI(TaskOperations.class) != null;
    } catch (OrchestraGwtException e) {
      return false;
    }
  }

  public List<MTask> getTasks() throws OrchestraGwtException {
    List<TTask> tasks;
    try {
      tasks = this.getAPI(TaskOperations.class).getMyTasks("ALL", null, null, null, null, null, null);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getTasks : \n" + Misc.getStackTraceFrom(e));
    }
    List<MTask> result = new ArrayList<MTask>();
    for (TTask task : tasks) {
      MTask mTask = new MTask(
          task.getTaskType(),
          task.getId(),
          task.getStatus().value(),
          task.getPresentationName(),
          task.getPresentationSubject(),
          task.getCreatedOn());
      result.add(mTask);
    }
    return result;
  }

  public void startTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).start(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - startTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void completeTask(String taskId, String data) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).complete(taskId, data);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - completeTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public MTaskDetails getTaskDetails(String taskId) throws OrchestraGwtException {
    try {
      TTask tTask = this.getAPI(TaskOperations.class).getTaskInfo(taskId);
      String description = this.getAPI(TaskOperations.class).getTaskDescription(taskId, null);
      List<String> potentialOwners = new ArrayList<String>();
      TOrganizationalEntity owners = tTask.getPotentialOwners();
      if (owners != null) {
        if (owners.getUsers() != null) {
          potentialOwners.addAll(owners.getUsers().getUser());
        }
        if (owners.getGroups() != null) {
          potentialOwners.addAll(owners.getGroups().getGroup());
        }
      }
      return new MTaskDetails(
          tTask.getTaskType(),
          tTask.getId(),
          tTask.getStatus().value(),
          tTask.getPresentationName(),
          tTask.getPresentationSubject(),
          tTask.getCreatedOn(),
          description,
          tTask.getActualOwner(),
          potentialOwners
      );
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - getTaskDetails : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void claimTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).claim(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - claimTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void stopTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).stop(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - stopTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void suspendTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).suspend(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - suspendTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void releaseTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).release(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - releaseTask : \n" + Misc.getStackTraceFrom(e));
    }
  }

  public void resumeTask(String taskId) throws OrchestraGwtException {
    try {
      this.getAPI(TaskOperations.class).resume(taskId);
    } catch (Exception e) {
      throw new OrchestraGwtException("Error while sending request to orchestra server !",
          "OrchestraManagement - resumeTask : \n" + Misc.getStackTraceFrom(e));
    }
  }
}
